from base_st_classes import *
from nfs4packer import NFS4Packer

class CompoundSuite(NFSSuite):
    """Test COMPOUND procedure

    Equivalence partitioning:

    Input Condition: tag
        Valid equivalence classes:
            no tag (0)
            tag (1)
        Invalid equivalence classes:
            -
    Input Condition: minorversion
        Valid equivalence classes:
            supported minorversions(2)
        Invalid equivalence classes:
            unsupported minorversions(3)
    Input Condition: argarray
        Valid equivalence classes:
            valid operations array(4)
        Invalid equivalence classes:
            invalid operations array(5)

    """

    #
    # Testcases covering valid equivalence classes.
    #
    def testZeroOps(self):
        """COMPOUND without operations

        Covered valid equivalence classes: 0, 2, 4
        """
        res = self.ncl.do_ops([])
        self.assert_OK(res)

    def testWithTag(self):
        """COMPOUND with tag

        Covered valid equivalence classes: 1, 2, 4
        """
        res = self.ncl.do_ops([self.ncl.putrootfh_op()],
                              tag="nfs4st.py test tag")
        self.assert_OK(res)

    #
    # Testcases covering invalid equivalence classes.
    #
    def testInvalidMinor(self):
        """COMPOUND with invalid minor version

        Covered invalid equivalence classes: 3

        Comments: Also verifies that the result array after
        NFS4ERR_MINOR_VERS_MISMATCH is empty.

        """
        res = self.ncl.do_ops([self.ncl.putrootfh_op()], minorversion=0xFFFF)
        self.assert_status(res, [NFS4ERR_MINOR_VERS_MISMATCH])

        self.failIf(res.resarray, "expected empty result array after"\
                    "NFS4ERR_MINOR_VERS_MISMATCH")

    def _verify_notsupp(self, opnum):
        """COMPOUND with undefined operations"""

        # pack_nfs_argop4 does not allow packing invalid operations.
        class custom_packer(NFS4Packer):
            def pack_nfs_argop4(self, data):
                self.pack_nfs_opnum4(data.argop)

        ncl = self.create_client(self.ncl.uid, self.ncl.gid, "badcompound")
        ncl.packer = custom_packer()

        op = nfs4types.nfs_argop4(argop=opnum)

        try:
            # This *should* raise the BadCompoundRes exception.
            res =  ncl.compound([op])
            # Ouch. This should not happen.
            # MARTIN - actually, it could, see following except clause.
            if len(res.resarray) == 0:
                self.fail("Expected OP_ILLEGAL resop4, argop list was empty.")
            opnum = res.resarray[0].resop
            self.failIf(opnum != OP_ILLEGAL, "Expected OP_ILLEGAL resop4, got %s" % nfs_opnum4_id[opnum])
            # self.fail("Expected BadCompoundRes exception. INTERNAL ERROR.")
        except BadDiscriminant, e:
            # MARTIN - this is flawed, because OP_ILLEGAL is a valid discriminant
            #          in a argop4 union. 
            # This should happen.
            self.failIf(e.value != OP_ILLEGAL,
                        "Expected result array with opnum %d, got %d" % (OP_ILLEGAL, e.value))
            # We have to do things a little by hand here
            pos = ncl.unpacker.get_position()
            data = ncl.unpacker.get_buffer()[pos:]
            errcode = nfs4lib.opaque2long(data)
            self.failIf(errcode != NFS4ERR_OP_ILLEGAL,
                        "Expected NFS4ERR_NOTSUPP, got %d" % errcode)
        except nfs4lib.InvalidCompoundRes, e:
            # FRED - must pull off end int and check vs valid_opnum_fn
            # print e
            self.fail(e)
        except rpc.RPCException, e:
            self.fail(e)

    def testUndefinedOps(self):
        """COMPOUND with operations 0, 1, 2 and 200 should return NFS4ERR_OP_ILLEGAL

        Covered invalid equivalence classes: 5

        Comments: The server should return NFS4ERR_OP_ILLEGAL for the
        undefined operations 0, 1 and 2. Although operation 2 may be
        introduced in later minor versions, the server should always
        return NFS4ERR_NOTSUPP if the minorversion is 0.
        """
        self._verify_notsupp(OP_ILLEGAL)
        self._verify_notsupp(0)
        self._verify_notsupp(1)
        self._verify_notsupp(2)
        self._verify_notsupp(200)


