from base_st_classes import *

class LockSuite(NFSSuite):
    """Test operation 12: LOCK

    """

    def setUp(self):
        NFSSuite.setUp(self)
        self.fh = None
        self.stateid = None
        self.init_connection()

    def tearDown(self):
        self.ncl.close_file(self.fh, self.stateid, error=None)
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass


    #
    # Tests that should work
    #
    def testFile(self):
        """LOCK a regular file
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_file(self.fh, self.stateid)
        self.ncl.lock_test(self.fh)

    def testClose(self):
        """LOCK - closing should release locks

        (FIXME - or it can return NFS4ERR_LOCKS_HELD, but right now LOTS
        of test would fail if this were true)
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_file(self.fh, self.stateid)
        self.ncl.lock_test(self.fh)
        self.ncl.close_file(self.fh, self.stateid)
        
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_test(self.fh, 0, 0xffffffffffffffff, error=[NFS4_OK])
        self.ncl.lock_file(self.fh, self.stateid, owner="testClose")
        self.ncl.lock_test(self.fh)

    def xxxtestExistingFile(self):
        """LOCK a regular file
        """
        dir = self.regfile[0:-1]
        name = self.regfile[-1]
        self.fh, self.stateid = self.ncl.open_confirm(dir, name)
        self.ncl.lock_file(self.fh, self.stateid)
        self.ncl.lock_test(self.fh)

    #
    # Tests that may or may not work depending on server
    #
    def test32bitRange(self):
        """LOCK range over 32 bits should work or return NFS4ERR_BAD_RANGE
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_file(self.fh, self.stateid, len=0xffffffffffff,
                           error=[NFS4_OK,NFS4ERR_BAD_RANGE])

    def testOverlap(self):
        """LOCK: overlapping locks should work or return NFS4ERR_LOCK_RANGE
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 25, 75)
        
        # Lock again with overlapping lock
        lockid = self.ncl.relock_file(self.fh, lockid, 1, 50, 75,
                           error=[NFS4_OK, NFS4ERR_LOCK_RANGE])
        if not lockid: return
        
        # Test the lock
        res = self.ncl.lock_test(self.fh)

        # Make sure we now have a single lock from 25 to 125
        obj = res.resarray[-1].arm.arm
        self.assertEqual(obj.offset,25,"Offset not 25")
        self.assertEqual(obj.length,100,"Length not 100")
 
    def testDowngrade(self):
        """LOCK change from write to a read lock
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 25, 75)

        # Lock again with read lock
        lockid = self.ncl.relock_file(self.fh, lockid, 1, 25, 75, type=READ_LT,
                                      error=[NFS4_OK, NFS4ERR_LOCK_NOTSUPP])
        if not lockid: return

        # Test the lock
        res = self.ncl.lock_test(self.fh)

        # Make sure we now have a single lock from 25 to 125
        obj = res.resarray[-1].arm.arm
        self.assertEqual(obj.locktype,READ_LT,"Locktype not changed")

    def testMode(self):
        """LOCK: try to write-lock a read-mode file
        """
        self.fh, self.stateid = self.ncl.create_confirm(access=OPEN4_SHARE_ACCESS_READ)
        self.ncl.lock_file(self.fh, self.stateid,
                           error=[NFS4_OK, NFS4ERR_OPENMODE])

    #
    # Tests for various error conditions
    #
    
    def testZeroLen(self):
        """LOCK with len=0 should return NFS4ERR_INVAL
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_file(self.fh, self.stateid, len=0,error=[NFS4ERR_INVAL])

    def testLenTooLong(self):
        """LOCK with offset+len overflow should return NFS4ERR_INVAL
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_file(self.fh, self.stateid, 100, 0xfffffffffffffffe,
                           error=[NFS4ERR_INVAL])

    def testNoCfh(self):
        """LOCK with no (cfh) should return NFS4ERR_NOFILEHANDLE
        """
        self.fh, self.stateid = self.ncl.create_confirm()

        # Lock the file
        lock_owner = lock_owner4(self.ncl.clientid, "owner")
        open_owner = open_to_lock_owner4(self.ncl.get_seqid(),
                                         self.stateid, 0, lock_owner)
        locker = locker4(TRUE, open_owner=open_owner)
        lockop = self.ncl.lock_op(WRITE_LT, FALSE, 0, 10, locker)
        res = self.ncl.do_ops([lockop])
        self.assert_status(res,[NFS4ERR_NOFILEHANDLE])

    def testBadSeqid(self):
        """LOCK with a bad sequence should return NFS4ERR_BAD_SEQID
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 50, 50)
        
        # Lock again with incorrect seqid 
        self.ncl.relock_file(self.fh, lockid, 2, 0, 25,
                             error=[NFS4ERR_BAD_SEQID])

    def testOldStateid(self):
        """LOCK with old stateid should return NFS4ERR_OLD_STATEID
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 50, 50)
        self.ncl.relock_file(self.fh, lockid, 1, 0, 25)
        
        # Another lock using old lock_stateid
        self.ncl.relock_file(self.fh, lockid, 2, 30, 10,
                             error=[NFS4ERR_OLD_STATEID])

    def testStaleClientid(self):
        """LOCK with a bad clientid should return NFS4ERR_STALE_CLIENTID
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        
        # Lock the file
        ops = [self.ncl.putfh_op(self.fh)]
        lock_owner = lock_owner4(self.get_invalid_clientid(), "owner")
        open_owner = open_to_lock_owner4(self.ncl.get_seqid(),
                                         self.stateid, 0, lock_owner)
        locker = locker4(TRUE, open_owner=open_owner)
        lockop = self.ncl.lock_op(WRITE_LT, FALSE, 0, 50, locker)
        ops.append(lockop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res,[NFS4ERR_STALE_CLIENTID])
        
    def testInconsistentClientid(self):
        """LOCK with a inconsistent clientids should return NFS4ERR_BAD_STATEID?
        """
        self.init_connection(ncl=self.ncl2)
        self.fh, self.stateid = self.ncl.create_confirm()
        
        # Lock the file
        ops = [self.ncl.putfh_op(self.fh)]
        lock_owner = lock_owner4(self.ncl2.clientid, "owner")
        open_owner = open_to_lock_owner4(self.ncl.get_seqid(),
                                         self.stateid, 0, lock_owner)
        locker = locker4(TRUE, open_owner=open_owner)
        lockop = self.ncl.lock_op(WRITE_LT, FALSE, 0, 50, locker)
        ops.append(lockop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res,[NFS4ERR_BAD_STATEID])
        
    def testBadStateid(self):
        """LOCK should return NFS4ERR_BAD_STATEID if use a bad id
        """
        self.fh, self.stateid = self.ncl.create_confirm()
        
        # Lock the file using all 0 stateid
        stateid = stateid4(0, "")
        self.ncl.lock_file(self.fh, stateid, error=[NFS4ERR_BAD_STATEID])

    def testBadLockSeqid(self):
        """LOCK with newlockowner must set lockid to 0"""
        self.fh, self.stateid = self.ncl.create_confirm()
        ops = [self.ncl.putfh_op(self.fh)]
        lock_owner = lock_owner4(self.ncl.clientid, "owner")
        open_owner = open_to_lock_owner4(self.ncl.get_seqid(),
                                         self.stateid, 10, lock_owner)
        locker = locker4(TRUE, open_owner=open_owner)
        lockop = self.ncl.lock_op(WRITE_LT, FALSE, 0, 50, locker)
        ops.append(lockop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res,[NFS4ERR_BAD_SEQID])

    def testLockowner(self):
        """LOCK owner should not work after openowner closes file"""
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid = self.ncl.lock_file(self.fh, self.stateid, 25, 75)

        # Close file
        self.ncl.close_file(self.fh, self.stateid)

        # Attempt to keep using lockowner
        lockid = self.ncl.unlock_file(self.fh, lockid, 1, 25, 75,
                                      error=[NFS4ERR_BAD_STATEID])

    def testLockowner2(self):
        """LOCK owner should not work if reused with 2nd file"""
        self.fh, self.stateid = self.ncl.create_confirm(owner="LOCK2")
        lockid = self.ncl.lock_file(self.fh, self.stateid, 25, 75)
        self.fh2, self.stateid2 = self.ncl.create_confirm(name='foo',
                                                          owner="LOCK2")
        lockid2 = self.ncl.lock_file(self.fh2, self.stateid2, 25, 75,
                                     error=[NFS4ERR_BAD_STATEID])

    def testLockowner3(self):
        """LOCK owner with same name as openowner"""
        self.fh, self.stateid = self.ncl.create_confirm(owner="LOCK3")
        lockid = self.ncl.lock_file(self.fh, self.stateid, 25, 75, owner="LOCK3")

        # Attempt to keep using lockowner
        lockid = self.ncl.unlock_file(self.fh, lockid, 1, 25, 75)

        # Close file
        self.ncl.close_file(self.fh, self.stateid)

    def testLockowner4(self):
        """LOCK owner created twice on same file should fail"""
        self.fh, self.stateid = self.ncl.create_confirm(owner="Lockowner4")
        lockid1 = self.ncl.lock_file(self.fh, self.stateid, 25, 75, owner="LOCK4")
        self.fh, self.stateid = self.ncl.open_confirm(owner="Lockowner4")
        lockid2 = self.ncl.lock_file(self.fh, self.stateid, 150, 75, owner="LOCK4", error=[NFS4ERR_BAD_STATEID])

    def testLockowner5(self):
        """LOCK owner created twice on two opens of same file should fail"""
        self.fh, self.stateid = self.ncl.create_confirm()
        lockid1 = self.ncl.lock_file(self.fh, self.stateid, 25, 75, owner="LOCK5")
        self.fh2, self.stateid2 = self.ncl.open_confirm()
        lockid2 = self.ncl.lock_file(self.fh2, self.stateid2, 150, 75, owner="LOCK5", error=[NFS4ERR_BAD_STATEID])

    def testRepeatedLock(self):
        """LOCK a regular file twice using newowner should fail"""
        self.fh, self.stateid = self.ncl.create_confirm()
        self.ncl.lock_file(self.fh, self.stateid)
        self.ncl.lock_test(self.fh)
        self.ncl.lock_file(self.fh, self.stateid, error=[NFS4ERR_BAD_STATEID])

