from base_st_classes import *

class ReaddirSuite(NFSSuite):
    """Test operation 26: READDIR

    FIXME: More testing of dircount/maxcount combinations.
    Note: maxcount represents READDIR4resok. Test this.
    fattr4_rdattr_error vs. global error

    Equivalence partitioning:
        
    Input Condition: current filehandle
        Valid equivalence classes:
            dir(10)
        Invalid equivalence classes:
            not dir(11)
            no filehandle(12)
    Input Condition: cookie
        Valid equivalence classes:
            zero(20)
            nonzero valid cookie(21)
        Invalid equivalence classes:
            invalid cookie(22)
    Input Condition: cookieverf
        Valid equivalence classes:
            zero(30)
            nonzero valid verifier(31)
        Invalid equivalence classes:
            invalid verifier(32)
    Input Condition: dircount
        Valid equivalence classes:
            zero(40)
            nonzero(41)
        Invalid equivalence classes:
            -
    Input Condition: maxcount
        Valid equivalence classes:
            nonzero(50)
        Invalid equivalence classes:
            zero(51)
    Input Condition: attrbits
        Valid equivalence classes:
            all requests without FATTR4_*_SET (60)
        Invalid equivalence classes:
            requests with FATTR4_*_SET (61)

    Comments: It's not possible to cover eq. class 22, since the cookie is
    opaque to the client. 
    """

    def tearDown(self):
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    #
    # Testcases covering valid equivalence classes.
    #
    def testFirst(self):
        """READDIR with cookie=0, maxcount=4096

        Covered valid equivalence classes: 10, 20, 30, 40, 50, 60
        """        
        lookupops = self.ncl.lookup_path(self.basedir)
        ops = [self.ncl.putrootfh_op()] + lookupops
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=4096,
                                        attr_request=[])
        ops.append(readdirop)
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Check that correct names are returned
        entries = []
        reply = res.resarray[-1].arm.arm.reply
        entry = reply.entries[0]
        while 1:
            entries.append(entry)
            if not entry.nextentry:
                break
            entry = entry.nextentry[0]
        name = [entry.name for entry in entries]
        correct_names = ["dev","doc","private","src","tmp"]
        self.assertEqual(name.sort(),correct_names.sort(),
                         "Dir not read correctly")
        
    def testAttr(self):
        """READDIR with attributes
        """        
        lookupops = self.ncl.lookup_path(self.basedir)
        ops = [self.ncl.putrootfh_op()] + lookupops
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=4096,
                                        attr_request=nfs4lib.list2attrmask([FATTR4_SIZE, FATTR4_FILEID]))
        ops.append(readdirop)
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Check that correct names are returned
        entries = []
        reply = res.resarray[-1].arm.arm.reply
        entry = reply.entries[0]
        while 1:
            entries.append(entry)
            if not entry.nextentry:
                break
            entry = entry.nextentry[0]
        name = [entry.name for entry in entries]
        correct_names = ["dev","doc","private","src","tmp"]
        self.assertEqual(name.sort(),correct_names.sort(),
                         "Dir not read correctly")
        #print nfs4lib.fattr2dict(entries[0].attrs)
        for e in entries:
            #print nfs4lib.fattr2dict(e.attrs)
            pass
        
    def testSubsequent(self):
        """READDIR with cookie from previous call

        Covered valid equivalence classes: 10, 21, 31, 41, 50, 60
        """
        # Call READDIR with small maxcount, to make sure not all
        # entries are returned.
        lookupops = self.ncl.lookup_path(self.basedir)
        ops = [self.ncl.putrootfh_op()] + lookupops
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=150,
                                        attr_request=[])
        ops.append(readdirop)
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Pull off returned cookie and verifier
        obj = res.resarray[-1].arm.arm
        cookieverf = obj.cookieverf
        list = obj.reply.entries
        while list[0].nextentry :
            list = list[0].nextentry
        cookie = list[0].cookie

        # Call READDIR a second time with saved cookie.
        lookupops = self.ncl.lookup_path(self.basedir)
        ops = [self.ncl.putrootfh_op()] + lookupops
        readdirop = self.ncl.readdir_op(cookie=cookie, cookieverf=cookieverf,
                                        dircount=0, maxcount=4096,
                                        attr_request=[])
        ops.append(readdirop)
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

    #
    # Testcases covering invalid equivalence classes.
    #
    def testFhNotDir(self):
        """READDIR with non-dir (cfh) should give NFS4ERR_NOTDIR

        Covered invalid equivalence classes: 11
        """
        for path in [ self.regfile,
                      self.blockfile,
                      self.charfile,
                      self.linkfile,
                      self.socketfile,
                      self.fifofile ] :
            lookupops = self.ncl.lookup_path(path)
            ops = [self.ncl.putrootfh_op()] + lookupops
            readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                            dircount=0, maxcount=4096,
                                            attr_request=[])
            ops.append(readdirop)
            res = self.ncl.do_ops(ops)
            self.assert_status(res, [NFS4ERR_NOTDIR])

    def testNoFh(self):
        """READDIR without (cfh) should return NFS4ERR_NOFILEHANDLE

        Covered invalid equivalence classes: 12
        """
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=4096,
                                        attr_request=[])
        res = self.ncl.do_ops([readdirop])
        self.assert_status(res, [NFS4ERR_NOFILEHANDLE])

    def testInvalidCookieverf(self):
        """READDIR with invalid cookieverf should return NFS4ERR_BAD_COOKIE

        Covered invalid equivalence classes: 32
        """
        # FIXME: Implement rest of testcase, as soon as
        # CITI supports dircount/maxcount. 
        self.skip("(TEST NOT IMPLEMENTED)")

    def testMaxcountZero(self):
        """READDIR with maxcount=0 should return NFS4ERR_TOOSMALL
        
        Covered invalid equivalence classes: 51
        """
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=0,
                                        attr_request=[])
        res = self.ncl.do_ops([self.ncl.putrootfh_op(), readdirop])
        self.assert_status(res, [NFS4ERR_TOOSMALL])

    def testSmallMaxcount(self):
        """READDIR with small maxcount should return NFS4ERR_TOOSMALL
        
        Covered invalid equivalence classes: 51
        """
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=10,
                                        attr_request=[])
        res = self.ncl.do_ops([self.ncl.putrootfh_op(), readdirop])
        self.assert_status(res, [NFS4ERR_TOOSMALL])

    def testWriteOnlyAttributes(self):
        """READDIR with attrs=FATTR4_*_SET should return NFS4ERR_INVAL

        Covered invalid equivalence classes: 61

        Comments: See GetattrSuite.testWriteOnlyAttributes. 
        """
        for attr in [FATTR4_TIME_ACCESS_SET,FATTR4_TIME_MODIFY_SET]:
            attrmask = nfs4lib.list2attrmask([attr])
            readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                            dircount=0, maxcount=4096,
                                            attr_request=attrmask)
            res = self.ncl.do_ops([self.ncl.putrootfh_op(), readdirop])
            self.assert_status(res, [NFS4ERR_INVAL])
        

    #
    # Extra tests.
    #
    def testUnaccessibleDirWithAttrs(self):
        """READDIR with (cfh) in unaccessible directory and getting attrs

        Extra test
        """
        lookupops = self.ncl.lookup_path(self.notaccessibledir)
        ops = [self.ncl.putrootfh_op()] + lookupops

        attrmask = nfs4lib.list2attrmask([FATTR4_RDATTR_ERROR,FATTR4_TYPE, FATTR4_SIZE, FATTR4_TIME_MODIFY])
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=4096,
                                        attr_request=attrmask)
        ops.append(readdirop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_ACCESS])

    def testUnaccessibleDir(self):
        """READDIR with (cfh) in unaccessible directory

        Extra test
        """
        lookupops = self.ncl.lookup_path(self.notaccessibledir)
        ops = [self.ncl.putrootfh_op()] + lookupops

        attrmask = []
        readdirop = self.ncl.readdir_op(cookie=0, cookieverf="\x00",
                                        dircount=0, maxcount=4096,
                                        attr_request=attrmask)
        ops.append(readdirop)
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_ACCESS])

    def testDots(self):
        """READDIR should not return . and .. in /doc

        Extra test

        Although some servers may actually support files named "." and
        "..", no files named "." or ".." should exist in /doc.
        """
        # Lookup fh for /doc
        fh = self.do_rpc(self.ncl.do_getfh, self.dirfile)

        # Get entries
        entries = self.do_rpc(self.ncl.do_readdir, fh)
        names = [entry.name for entry in entries]

        self.failIf("." in names,
                    "READDIR in /doc returned .-entry")

        self.failIf(".." in names,
                    "READDIR in /doc returned ..-entry")

    def testStrangeNames(self):
        """READDIR should obey OPEN naming policy

        Extra test

        Comments: Verifying that readdir obeys the same naming policy
        as OPEN.
        """
        self.init_connection()
        
        try:
            (accepted_names, rejected_names) = self.try_file_names(remove_files=0)
        except SkipException, e:
            self.skip(e)

        fh = self.do_rpc(self.ncl.do_getfh, self.tmp_dir) 
        entries = self.do_rpc(self.ncl.do_readdir, fh)
        readdir_names = [entry.name for entry in entries]

        # Verify that READDIR returned all accepted_names
        missing_names = []
        for name in accepted_names:
            if name not in readdir_names:
                missing_names.append(name)

        self.failIf(missing_names, "Missing names in READDIR results: %s" \
                    % missing_names)

        # ... and nothing more
        extra_names = []
        for name in readdir_names:
            if not name in accepted_names:
                extra_names.append(name)

        self.failIf(extra_names, "Extra names in READDIR results: %s" \
                    % extra_names)

    def testReservedCookies(self):
        """READDIR with a reserved cookie should return NFS4ERR_BAD_COOKIE
        
        """
        for cookie in [1,2]:
            readdirop = self.ncl.readdir_op(cookie=cookie, cookieverf="\x00",
                                            dircount=0, maxcount=4096,
                                            attr_request=[])
            res = self.ncl.do_ops([self.ncl.putrootfh_op(), readdirop])
            self.assert_status(res, [NFS4ERR_BAD_COOKIE])

    def testBug(self):
        """READDIR had a strange bug - is it there?

        Extra test

        Comments: Verifying that readdir obeys the same naming policy
        as OPEN.
        """
        self.init_connection()
        
        try:
            (accepted_names, rejected_names) = self.small_try_file_names(remove_files=0)
        except SkipException, e:
            self.skip(e)

        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            self.fail(e)

    def small_try_file_names(self, remove_files=1, creator=None):
        """ Try all kinds of interesting file names and check if they are accepted
        (via accepted_filename). Creates files in self.tmp_dir. The optional argument
        remove_files can be unset to save created files. 

        Returns a tuple of (accepted_names, rejected_names)
        Raises SkipException if temporary directory cannot be cleared. You
        should catch it. 
        """
        if not creator:
            creator = self.create_via_open

        try_names = []
        rejected_names = []

        # aa<char>bb, with char 0-255
        for i in range(1, 256): #[1,2,3,255]
            if i in [47]: continue # 122
            ustr = "aa" + unichr(i) + "bb"
            try_names.append(ustr.encode("utf8"))

        # C:
        try_names.append("C:".encode("utf8"))

        #try_names.sort()

        # Ok, lets try all these names.
        # Begin with cleaning /tmp
        self.clean_dir(self.tmp_dir)

        for filename in try_names[:]:
            # Do not remove the file after creation
            if not self.accepted_filename(filename, remove_file=remove_files,
                                           creator=creator):
                try_names.remove(filename)
                rejected_names.append(filename)

        return (try_names, rejected_names)

